#   Copyright 2019 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_ising.py.

Contains the Ising class. See ``help(qubovert.Ising)``.

"""

from .utils import BO, IsingMatrix


__all__ = 'Ising',


class Ising(BO, IsingMatrix):
    """Ising.

    Class to manage converting general Ising problems to and from their
    QUBO and Ising formluations.

    This class deals with Isings that have binary labels that do not range from
    0 to n-1. If your labels are nonnegative integers, consider using
    ``qubovert.utils.IsingMatrix``. Note that it is generally
    more efficient to initialize an empty Ising object and then build the
    Ising, rather than initialize a Ising object with an already built dict.

    Ising inherits some methods and attributes the ``IsingMatrix`` class. See
    ``help(qubovert.utils.IsingMatrix)``.

    Ising inherits some methods and attributes the ``BO`` class. See
    ``help(qubovert.utils.BO)``.

    Example usage
    -------------
    >>> ising = Ising()
    >>> ising[('a',)] += 5
    >>> ising[(0, 'a')] -= 2
    >>> ising -= 1.5
    >>> ising
    {('a',): 5, ('a', 0): -2, (): -1.5}

    >>> ising = Ising({('a',): 5, (0, 'a'): -2, (): -1.5})
    >>> ising
    {('a',): 5, ('a', 0): -2, (): -1.5}
    >>> L = ising.to_ising()
    >>> L
    {(0,): 5, (0, 1): -2, (): -1.5}
    >>> ising.convert_solution({0: 1, 1: 0})
    {'a': 1, 0: 0}

    Note 1
    ------
    Note that keys will end up sorted by their hash. Hashes will not be
    consistent across Python sessions (unless they are integers)! For example,
    both of the following can happen:

    >>> print(Ising({('a', 0): 1, (0, 1): -1}))
    {('a', 0): 1, (0, 1): -1}

    or

    >>> print(Ising({('a', 0): 1, (0, 1): -1}))
    {(0, 'a'): 1, (0, 1): -1}

    But the following will never happen:

    >>> print(Ising({('a', 0): 1, (0, 1): -1}))
    {('a', 0): 1, (1, 0): -1}

    Ie integers will always be correctly sorted.

    Note 2
    ------
    For efficiency, many internal variables including mappings are computed as
    the problemis being built. This can cause these
    values to be wrong for some specific situations. Calling ``refresh``
    will rebuild the dictionary, resetting all of the values. See
    ``help(Ising.refresh)``

    Examples
    --------
    >>> from qubovert import Ising
    >>> L = Ising()
    >>> L[('a',)] += 1
    >>> L, L.mapping, L.reverse_mapping
    {('a',): 1}, {'a': 0}, {0: 'a'}
    >>> L[('a',)] -= 1
    >>> L, L.mapping, L.reverse_mapping
    {}, {'a': 0}, {0: 'a'}
    >>> L.refresh()
    >>> L, L.mapping, L.reverse_mapping
    {}, {}, {}

    """

    def __init__(self, *args, **kwargs):
        """__init__.

        This class deals with Isings that have binary labels that do not range
        from 0 to n-1. If your labels are nonnegative integers, consider using
        ``qubovert.utils.IsingMatrix``. Note that it is generally more
        efficient to initialize an empty Ising object and then build the Ising,
        rather than initialize a Ising object with an already built dict.

        Parameters
        ----------
        argguments : define a dictionary with ``dict(*args, **kwargs)``.
            The dictionary will be initialized to follow all the convensions of
            the class.

        Examples
        -------
        >>> ising = Ising()
        >>> ising[('a',)] += 5
        >>> ising[(0, 'a')] -= 2
        >>> ising -= 1.5
        >>> ising
        {('a',): 5, ('a', 0): -2, (): -1.5}

        >>> ising = Ising({('a',): 5, (0, 'a'): -2, (): -1.5})
        >>> ising
        {('a',): 5, ('a', 0): -2, (): -1.5}

        """
        BO.__init__(self, *args, **kwargs)
        IsingMatrix.__init__(self, *args, **kwargs)

    def to_ising(self):
        """to_ising.

        Create and return upper triangular Ising representing the problem.
        The labels will be integers from 0 to n-1.

        Return
        -------
        L : qubovert.utils.IsingMatrix object.
            The upper triangular Ising matrix, a IsingMatrix object.
            For most practical purposes, you can use IsingMatrix in the
            same way as an ordinary dictionary. For more information,
            see ``help(qubovert.utils.IsingMatrix)``.

        """
        L = IsingMatrix()

        for k, v in self.items():
            key = tuple(self._mapping[i] for i in k)
            L[key] += v

        return L

    def convert_solution(self, solution):
        """convert_solution.

        Convert the solution to the integer labeled Ising to the solution to
        the originally labeled Ising.

        Parameters
        ----------
        solution : iterable or dict.
            The QUBO or Ising solution output. The Ising solution output
            is either a list or tuple where indices specify the label of the
            variable and the element specifies whether it's 0 or 1 for QUBO
            (or -1 or 1 for Ising), or it can be a dictionary that maps the
            label of the variable to is value.

        Return
        -------
        res : dict.
            Maps binary variable labels to their Ising solutions values {0, 1}.

        Example
        -------
        >>> ising = Ising({('a',): 1, ('a', 'b'): -2, ('c',): -1})
        >>> L = ising.to_ising()
        >>> L
        {(0,): 1, (0, 1): -2, (2,): -1}
        >>> solution = solve_ising(L)  # any solver you want
        >>> solution
        [1, 1, -1]  # or {0: 1, 1: 1, 2: -1}
        >>> sol = ising.convert_solution(solution)
        >>> sol
        {'a': 1, 'b': 1, 'c': -1}

        >>> ising = Ising({('a',): 1, ('a', 'b'): -2, ('c',): -1})
        >>> L = ising.to_qubo()
        >>> solution = solve_ising(L)  # any solver you want
        >>> solution
        [1, 1, 0]  # or {0: 1, 1: 1, 2: 0}
        >>> sol = ising.convert_solution(solution)
        >>> sol
        {'a': 1, 'b': 1, 'c': -1}

        """
        return {
            self._reverse_mapping[i]: 1 if solution[i] == 1 else -1
            for i in range(self.num_binary_variables)
        }

    @staticmethod
    def _check_key_valid(key):
        """_check_key_valid.

        Internal method to check if an input key to the dictionary is valid.
        Checks to see if ``key`` is a tuple.

        Parameters
        ---------
        key : anything, but must be a tuple to be valid.

        Returns
        -------
        None.

        Raises
        ------
        KeyError if the key is invalid.

        """
        # override IsingMatrix._check_key_valid to allow for noninteger keys.
        invalid = (
            not isinstance(key, tuple) or
            len(set(x for x in set(key) if key.count(x) % 2)) > 2
        )
        if invalid:
            raise KeyError(
                "Key formatted incorrectly, must be tuple of <= 2 unique "
                "element. See HIsing for arbitrary number of unique elements.")



#   Copyright 2019 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_hising.py.

Contains the HIsing class. See ``help(qubovert.HIsing)``.

"""

from .utils import BO, HIsingMatrix, hising_to_pubo
from . import Ising


__all__ = 'HIsing',


class HIsing(BO, HIsingMatrix):
    """HIsing.

    Class to manage converting general HIsing problems to and from their
    HIing, PUBO, Ising, and QUBO formluations. In general, this class
    deals with unconstrained optimization problems that have arbitrary degree.
    To convert this to a Ising (see ``to_ising``) or QUBO (``to_qubo``) we have
    to introduce ancilla variables. The ``convert_solution`` method deals with
    converting a solution to the problem with ancilla variables back to the
    solution to the original problem.

    This class deals with HIsings that have binary labels that do not range
    from 0 to n-1. Note that it is generally
    more efficient to initialize an empty HIsing object and then build the
    HIsing, rather than initialize a HIsing object with an already built dict.

    HIsing inherits some methods and attributes the ``HIsingMatrix`` class. See
    ``help(qubovert.utils.HIsingMatrix)``.

    HIsing inherits some methods and attributes the ``BO`` class. See
    ``help(qubovert.utils.BO)``.

    Example usage
    -------------
    >>> hising = HIsing()
    >>> hising[('a',)] += 5
    >>> hising[(0, 'a', 1)] -= 2
    >>> hising -= 1.5
    >>> hising
    {('a',): 5, ('a', 0, 1): -2, (): -1.5}

    >>> hising = HIsing({('a',): 5, (0, 'a', 1): -2, (): -1.5})
    >>> hising
    {('a',): 5, ('a', 0, 1): -2, (): -1.5}
    >>> H = hising.to_hising()
    >>> H
    {(0,): 5, (0, 1): -2, (): -1.5}
    >>> hising.convert_solution({0: 1, 1: -1, 2: 1})
    {'a': 1, 0: -1, 1: 1}

    In the next example, notice that we introduce ancilla variables to
    represent that ```(0, 1)`` term. See the ``to_ising`` method for more info.

    >>> hising = HIsing({('a',): 5, (0, 'a', 1): -2, (): -1.5})
    >>> hising.mapping
    {'a': 0, 0: 1, 1: 2}
    >>> L = hising.to_ising()
    >>> L
    {(0,): 0.75, (): 11.25, (3,): 8.5, (0, 1): 4.25, (1,): -4.25, (0, 3): -8.5,
     (1, 3): -8.5, (2, 3): -4.0, (1, 2): 2.0, (0, 2): 2.0, (2,): -2.0}
    >>> hising.convert_solution({0: 1, 1: -1, 2: 1, 3: -1})
    {'a': 1, 0: -1, 1: 1}

    Note 1
    ------
    Note that keys will end up sorted by their hash. Hashes will not be
    consistent across Python sessions (unless they are integers)! For example,
    both of the following can happen:

    >>> print(HIsing({('a', 0): 1, (0, 1): -1}))
    {('a', 0): 1, (0, 1): -1}

    or

    >>> print(HIsing({('a', 0): 1, (0, 1): -1}))
    {(0, 'a'): 1, (0, 1): -1}

    But the following will never happen:

    >>> print(HIsing({('a', 0): 1, (0, 1): -1}))
    {('a', 0): 1, (1, 0): -1}

    Ie integers will always be correctly sorted.

    Note 2
    ------
    For efficiency, many internal variables including mappings are computed as
    the problemis being built. This can cause these
    values to be wrong for some specific situations. Calling ``refresh``
    will rebuild the dictionary, resetting all of the values. See
    ``help(HIsing.refresh)``

    Examples
    --------
    >>> from qubovert import HIsing
    >>> H = HIsing()
    >>> H[('a',)] += 1
    >>> H, H.mapping, H.reverse_mapping
    {('a',): 1}, {'a': 0}, {0: 'a'}
    >>> H[('a',)] -= 1
    >>> H, H.mapping, H.reverse_mapping
    {}, {'a': 0}, {0: 'a'}
    >>> H.refresh()
    >>> H, H.mapping, H.reverse_mapping
    {}, {}, {}

    """

    def __init__(self, *args, **kwargs):
        """__init__.

        This class deals with HIsings that have binary labels that do not range
        from 0 to n-1. Note that it is generally more efficient to initialize
        an empty HIsing object and then build the HIsing, rather than
        initialize a HIsing object with an already built dict.

        Parameters
        ----------
        arguments : define a dictionary with ``dict(*args, **kwargs)``.
            The dictionary will be initialized to follow all the convensions of
            the class.

        Examples
        -------
        >>> hising = HIsing()
        >>> hising[('a',)] += 5
        >>> hising[(0, 'a')] -= 2
        >>> hising -= 1.5
        >>> hising
        {('a',): 5, ('a', 0): -2, (): -1.5}

        >>> hising = HIsing({('a',): 5, (0, 'a', 1): -2, (): -1.5})
        >>> hising
        {('a',): 5, ('a', 0, 1): -2, (): -1.5}

        """
        BO.__init__(self, *args, **kwargs)
        HIsingMatrix.__init__(self, *args, **kwargs)

    def _to_hising(self):
        """to_hising.

        Internal helper method.

        Create and return upper triangular degree HIsing representing
        the problem. The labels will be integers from 0 to n-1.

        Return
        -------
        H : qubovert.utils.HIsingMatrix object.
            The upper triangular HIsing matrix, a HIsingMatrix object.
            For most practical purposes, you can use HIsingMatrix in the
            same way as an ordinary dictionary. For more information,
            see ``help(qubovert.utils.HIsingMatrix)``.

        """
        H = HIsingMatrix()
        for k, v in self.items():
            key = tuple(sorted(self._mapping[i] for i in k))
            H[key] += v
        return H

    def _create_pubo(self):
        """_create_pubo.

        Internal method to create the PUBO object from self.

        Returns
        -------
        P : qubovert.PUBO object.

        """
        P = hising_to_pubo(self)
        P._mapping = self.mapping
        P._reverse_mapping = self.reverse_mapping
        return P

    def to_pubo(self, deg=None, lam=None):
        """to_pubo.

        Create and return upper triangular degree ``deg`` PUBO representing the
        HIsing problem. The labels will be integers from 0 to n-1.

        Parameters
        ----------
        deg : int >= 0 (optional, defaults to None).
            The degree of the final PUBO. If ``deg`` is None, then the degree
            of the output PUBO will be the same as the degree of ``self``,
            ie see ``self.degree``.
        lam : function (optional, defaults to None).
            Note that if ``deg`` is None or ``deg >= self.degree``, then
            ``lam`` is unneccessary and will not be used.
            If ``lam`` is None, the function ``PUBO.default_lam`` will be used.
            ``lam`` is the penalty factor to introduce in order to enforce the
            ancilla constraints. When we reduce the degree of the model, we add
            penalties to the lower order model in order to enforce ancilla
            variable constraints. These constraints will be multiplied by
            ``lam(v)``, where ``v`` is the value associated with the term that
            it is reducing. For example, a term ``(0, 1, 2): 3`` in the higher
            order model may be reduced to a term ``(0, 3): 3`` for the lower
            order model, and then the fact that ``3`` should be the product
            of ``1`` and ``2`` will be enforced with a penalty weight
            ``lam(3)``.

        Return
        -------
        P : qubovert.utils.PUBOMatrix object.
            The upper triangular PUBO matrix, a PUBOMatrix object.
            For most practical purposes, you can use PUBOMatrix in the
            same way as an ordinary dictionary. For more information,
            see ``help(qubovert.utils.PUBOMatrix)``.

        Notes
        -----
        The penalty that we use to enforce the constraints that the ancilla
        variable ``z`` is equal to the product of the two variables that it is
        replacing, ``xy``, is:

            ``0`` if ``z == xy``,
            ``3*lam(v)`` if ``x == y == 0 and z == 1``, and
            ``lam(v)`` else.

        See https://arxiv.org/pdf/1307.8041.pdf equation 6.

        """
        return self._create_pubo().to_pubo(deg, lam)

    def to_hising(self, deg=None, lam=None):
        """to_hising.

        Create and return upper triangular degree ``deg`` HIsing representing
        the problem. The labels will be integers from 0 to n-1.

        Parameters
        ----------
        deg : int >= 0 (optional, defaults to None).
            The degree of the final HIsing. If ``deg`` is None, then the degree
            of the output HIsing will be the same as the degree of ``self``,
            ie see ``self.degree``.
        lam : function (optional, defaults to None).
            Note that if ``deg`` is None or ``deg >= self.degree``, then
            ``lam`` is unneccessary and will not be used.
            If ``lam`` is None, the function ``PUBO.default_lam`` will be used.
            ``lam`` is the penalty factor to introduce in order to enforce the
            ancilla constraints. When we reduce the degree of the model, we add
            penalties to the lower order model in order to enforce ancilla
            variable constraints. These constraints will be multiplied by
            ``lam(v)``, where ``v`` is the value associated with the term that
            it is reducing. For example, a term ``(0, 1, 2): 3`` in the higher
            order model may be reduced to a term ``(0, 3): 3`` for the lower
            order model, and then the fact that ``3`` should be the product
            of ``1`` and ``2`` will be enforced with a penalty weight
            ``lam(3)``.

        Return
        -------
        H : qubovert.utils.HIsingMatrix object.
            The upper triangular HIsing matrix, a HIsingMatrix object.
            For most practical purposes, you can use HIsingMatrix in the
            same way as an ordinary dictionary. For more information,
            see ``help(qubovert.utils.HIsingMatrix)``.

        Notes
        -----
        See the ``to_pubo`` docstring for more information on the penalties and
        ``lam``.

        """
        if deg is None or deg >= self.degree:
            return self._to_hising()
        return self._create_pubo().to_hising(deg, lam)

    def to_qubo(self, lam=None):
        """to_qubo.

        Create and return upper triangular QUBO representing the problem.
        The labels will be integers from 0 to n-1. We introduce ancilla
        variables in order to reduce the degree of the HIsing to a QUBO. The
        solution to the HIsing can be read from the solution to the QUBO by
        using the ``convert_solution`` method.

        Parameters
        ----------
        lam : function (optional, defaults to None).
            If ``lam`` is None, the function ``PUBO.default_lam`` will be used.
            ``lam`` is the penalty factor to introduce in order to enforce the
            ancilla constraints. When we reduce the degree of the HIsing to a
            QUBO, we add penalties to the QUBO in order to enforce ancilla
            variable constraints. These constraints will be multiplied by
            ``lam(v)``, where ``v`` is the value associated with the term that
            it is reducing. For example, a term ``(0, 1, 2): 3`` in the Hising
            may be reduced to a term ``(0, 3): 3`` for the Ising, and then the
            fact that ``3`` should be the product of ``1`` and ``2`` will be
            enforced with a penalty weight ``lam(3)``.

        Return
        -------
        Q : qubovert.utils.QUBOMatrix object.
            The upper triangular QUBO matrix, an QUBOMatrix object.
            For most practical purposes, you can use QUBOMatrix in the
            same way as an ordinary dictionary. For more information,
            see ``help(qubovert.utils.QUBOMatrix)``.

        """
        return self._create_pubo().to_qubo(lam)

    def convert_solution(self, solution):
        """convert_solution.

        Convert the solution to the integer labeled HIsing to the solution to
        the originally labeled HIsing.

        Parameters
        ----------
        solution : iterable or dict.
            The HIsing, HIsing, Ising, or Ising solution output. The HIsing
            solution output is either a list or tuple where indices specify the
            label of the variable and the element specifies whether it's 0 or 1
            for HIsing (or -1 or 1 for Ising), or it can be a dictionary that
            maps the label of the variable to is value. The Ising/Ising
            solution output includes the assignment for the ancilla variables
            used to reduce the degree of the HIsing.

        Return
        -------
        res : dict.
            Maps binary variable labels to their HIsing solutions values
            {-1, 1}.

        Example
        -------
        >>> hising = HIsing({('a',): 5, (0, 'a', 1): -2, (): -1.5})
        >>> hising
        {('a',): 5, ('a', 0, 1): -2, (): -1.5}
        >>> H = hising.to_hising()
        >>> H
        {(0,): 5, (0, 1): -2, (): -1.5}
        >>> hising.convert_solution({0: 1, 1: -1, 2: 1})
        {'a': 1, 0: -1, 1: 1}

        In the next example, notice that we introduce ancilla variables to
        represent that ```(0, 1)`` term. See the ``to_ising`` method for more
        info.

        >>> hising = HIsing({('a',): 5, (0, 'a', 1): -2, (): -1.5})
        >>> hising.mapping
        {'a': 0, 0: 1, 1: 2}
        >>> L = hising.to_ising(3)
        >>> L
        {(0,): 5, (0, 3): -2, (): -2.25,
         (1, 2): 3/4, (2, 3): 3/4, (1, 3): 3/4, (1,): 3/2, (2,): 3/2, (3,): 3}
        >>> hising.convert_solution({0: 1, 1: -1, 2: 1, 2: -1})
        {'a': 1, 0: -1, 1: 1}

        Notes
        -----
        We take ignore the ancilla variable assignments when we convert the
        solution. For example if the conversion from HIsing to Ising introduced
        an ancilla varable ``z = xy`` where ``x`` and ``y`` are variables of
        the HIsing, then ``solution`` must have values for ``x``, ``y``, and
        ``z``. If the Ising solver found that ``x = 1``, ``y = -1``, and
        ``z = 1``, then the constraint that ``z = xy`` is not satisfied (one
        possible cause for this is if the ``lam`` argument in ``to_ising`` is
        too small). ``convert_solution`` will return that ``x = 1`` and
        ``y = -1`` and ignore the value of ``z``.

        """
        # this works for converting a solution to the pubo, qubo, hising, or
        # ising formulations, since in the to_ising function all ancilla
        # variables are labeled with integers >= self.num_binary_variables.
        return Ising.convert_solution(self, solution)

    @staticmethod
    def _check_key_valid(key):
        """_check_key_valid.

        Internal method to check if an input key to the dictionary is valid.
        Checks to see if ``key`` is a tuple.

        Parameters
        ---------
        key : anything, but must be a tuple to be valid.

        Returns
        -------
        None.

        Raises
        ------
        KeyError if the key is invalid.

        """
        # override HIsingMatrix._check_key_valid to allow for noninteger keys.
        if not isinstance(key, tuple):
            raise KeyError(
                "Key formatted incorrectly, must be tuple")



#   Copyright 2019 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_hobo.py.

Contains the HOBO class. See ``help(qubovert.HOBO)``.

"""

from . import PUBO
from .utils import QUBOVertWarning
from .sat import OR, XOR
from numpy import log2, ceil


__all__ = 'HOBO', 'binary_var'


# special constraint forms

def _special_constraints_le_zero(hobo, P, lam):
    """_special_constraints_le_zero.

    See if the constraint that ``P <= 0`` matches any special forms.

    Parameters
    ----------
    hobo : HOBO object.
        The HOBO to add the constraint to.
    P : PUBO object.
        The PUBO constraint such that ``P <= 0``.
        are present.
    lam : float > 0 or sympy.Symbol.
        Langrange multiplier to penalize violations of the constraint.

    Return
    ------
    success : bool.
        True if a special constraint was found and added to ``hobo``, else
        False.

    """
    # if P is of the form sum(x_i) <= 1.
    if P.offset == -1 and all(x == 1 for x in (P + 1).values()):
        hobo += lam * P * (P + 1) / 2
        return True

    return False


# helpers

def _create_tuple(x):
    """_create_tuple.

    Create a tuple from ``x``.

    Parameters
    ----------
    x : object or list.

    Return
    ------
    res : tuple.

    """
    return tuple(x) if isinstance(x, list) else (x,)


def _pubo_value_extrema(P):
    """_pubo_value_extrema.

    Find the approximate minimum and maximum possible values that a PUBO can
    take.

    Parameters
    ----------
    P : PUBO object.

    Return
    ------
    res : tuple (min, max).

    """
    offset = P.offset
    P -= offset
    min_, max_ = offset, offset
    for v in P.values():
        if v < 0:
            min_ += v
        elif v > 0:
            max_ += v
    P += offset
    return min_, max_


def _get_bounds(P, bounds):
    """_get_bounds.

    Compute the missing bounds of ``P``.

    Parameters
    ----------
    P : PUBO object.
    bounds : two element tuple or None.
        A tuple ``(min, max)``, the minimum and maximum values that the
        PUBO ``P`` can take. If ``bounds`` is None, then they will be
        calculated (approximately), or if either of the elements of
        ``bounds`` is None, then that element will be calculated
        (approximately).

    Return
    ------
    res : two element tuple.
        The bounds of ``P``.

    """
    if bounds is None or bounds == (None, None):
        bounds = _pubo_value_extrema(P)
    elif bounds[0] is None:
        bounds = _pubo_value_extrema(P)[0], bounds[1]
    elif bounds[1] is None:
        bounds = bounds[0], _pubo_value_extrema(P)[1]

    return bounds


def binary_var(name):
    """binary_var.

    Create a HOBO (see ``qubovert.HOBO``) from a single binary variable.

    Parameters
    ----------
    name : any hashable object.
        Name of the binary variable.

    Return
    ------
    hobo : qubovert.HOBO object.
        The model representing the binary variable.

    Examples
    --------
    >>> from qubovert import binary_var, HOBO
    >>>
    >>> x0 = binary_var("x0")
    >>> print(x0)
    {('x0',): 1}
    >>> print(isinstance(x0, HOBO))
    True

    >>> x = [binary_var('x{}'.format(i)) for i in range(5)]
    >>> hobo = sum(x)
    >>> print(hobo)
    {('x0',): 1, ('x1',): 1, ('x2',): 1, ('x3',): 1, ('x4',): 1}
    >>> hobo **= 2
    >>> print(hobo)
    {('x0',): 1, ('x0', 'x1'): 2, ('x2', 'x0'): 2, ('x3', 'x0'): 2,
     ('x4', 'x0'): 2, ('x1',): 1, ('x2', 'x1'): 2, ('x3', 'x1'): 2,
     ('x4', 'x1'): 2, ('x2',): 1, ('x2', 'x3'): 2, ('x2', 'x4'): 2, ('x3',): 1,
     ('x4', 'x3'): 2, ('x4',): 1}
    >>> hobo *= -1
    >>> print(hobo.solve_bruteforce())
    {'x0': 1, 'x1': 1, 'x2': 1, 'x3': 1, 'x4': 1}
    >>> hobo.add_constraint_eq_zero(x[0] + x[1])
    >>> print(hobo.solve_bruteforce())
    {'x0': 0, 'x1': 0, 'x2': 1, 'x3': 1, 'x4': 1}

    """
    return HOBO({(name,): 1})


# main class

class HOBO(PUBO):
    """HOBO.

    This class deals with Higher Order Binary Optimization problems. HOBO
    inherits some methods and attributes from the ``PUBO`` class. See
    ``help(qubovert.PUBO)``.

    ``HOBO`` has all the same methods as ``PUBO``, but adds some constraint
    methods; namely

    - ``add_constraint_eq_zero(P, lam=1, ...)`` enforces that ``P == 0`` by
      penalizing with ``lam``,
    - ``add_constraint_lt_zero(P, lam=1, ...)`` enforces that ``P < 0`` by
      penalizing with ``lam``,
    - ``add_constraint_le_zero(P, lam=1, ...)`` enforces that ``P <= 0`` by
      penalizing with ``lam``,
    - ``add_constraint_gt_zero(P, lam=1, ...)`` enforces that ``P > 0`` by
      penalizing with ``lam``, and
    - ``add_constraint_ge_zero(P, lam=1, ...)`` enforces that ``P >= 0`` by
      penalizing with ``lam``.

    Each of these takes in a PUBO ``P`` and a lagrange multiplier ``lam`` that
    defaults to 1. See each of their docstrings for important details on their
    implementation.

    We then implement logical operations:

    - ``add_constraint_AND``, ``add_constraint_NAND``,
    - ``add_constraint_eq_AND``, ``add_constraint_eq_NAND``,
    - ``add_constraint_OR``, ``add_constraint_NOR```,
    - ``add_constraint_eq_OR``, ``add_constraint_eq_NOR``,
    - ``add_constraint_XOR``, ``add_constraint_XNOR``,
    - ``add_constraint_eq_XOR``, ``add_constraint_eq_XNOR``,
    - ``add_constraint_ONE``, ``add_constraint_NOT``,
    - ``add_constraint_eq_ONE``, ``add_constraint_eq_NOT``.

    See each of their docstrings for important details on their implementation.

    Notes
    -----
    - Variables names that begin with ``"__a"`` should not be used since they
      are used internally to deal with some ancilla variables to enforce
      constraints.
    - The ``self.solve_bruteforce`` method will solve the HOBO ensuring that
      all the inputted constraints are satisfied. Whereas
      ``qubovert.utils.solve_pubo_bruteforce(self)`` or
      ``qubovert.utils.solve_pubo_bruteforce(self.to_pubo())`` will solve the
      PUBO created from the HOBO. If the inputted constraints are not enforced
      strong enough (ie too small lagrange multipliers) then these may not give
      the correct result, whereas ``self.solve_bruteforce()`` will always give
      the correct result (ie one that satisfies all the constraints).

    Examples
    --------
    See ``qubovert.PUBO`` for more examples of using HOBO without constraints.

    >>> H = HOBO()
    >>> H.add_constraint_eq_zero({('a', 1): 2, (1, 2): -1, (): -1})
    >>> H
    {('a', 1, 2): -4, (1, 2): 3, (): 1}
    >>> H -= 1
    >>> H
    {('a', 1, 2): -4, (1, 2): 3}

    >>> H = HOBO()
    >>>
    >>> # minimize -x_0 - x_1 - x_2
    >>> for i in (0, 1, 2):
    >>>     H[(i,)] -= 1
    >>>
    >>> # subject to constraints
    >>> H.add_constraint_eq_zero(  # enforce that x_0 x_1 - x_2 == 0
    >>>     {(0, 1): 1, (2,): -1}
    >>> ).add_constraint_lt_zero(  # enforce that x_1 x_2 + x_0 < 1
    >>>     {(1, 2): 1, (0,): 1, (): -1}
    >>> )
    >>> print(H)
    >>> # {(1,): -2, (2,): -1, (0, 1): 2, (1, 2): 2, (0, 1, 2): 2}
    >>>
    >>> print(H.solve_bruteforce(all_solutions=True))
    >>> # [{0: 0, 1: 1, 2: 0}]
    >>>
    >>> Q = H.to_qubo()
    >>> solutions = [H.convert_solution(sol)
    >>>              for sol in Q.solve_bruteforce(all_solutions=True)]
    >>> print(solutions)
    >>> # [{0: 0, 1: 1, 2: 0}]  # matches the HOBO solution!
    >>>
    >>> L = H.to_ising()
    >>> solutions = [H.convert_solution(sol)
    >>>              for sol in L.solve_bruteforce(all_solutions=True)]
    >>> print(solutions)
    >>> # [{0: 0, 1: 1, 2: 0}]  # matches the HOBO solution!

    Enforce that c == a b

    >>> H = HOBO().add_constraint_eq_AND('c', 'a', 'b')
    >>> H
    {('c',): 3, ('b', 'a'): 1, ('c', 'a'): -2, ('c', 'b'): -2}

    >>> from any_module import qubo_solver
    >>> # or from qubovert.utils import solve_qubo_bruteforce as qubo_solver
    >>> H = HOBO()
    >>>
    >>> # make it favorable to AND variables a and b, and variables b and c
    >>> H.add_constraint_AND('a', 'b').add_constraint_AND('b', 'c')
    >>>
    >>> # make it favorable to OR variables b and c
    >>> H.add_constraint_OR('b', 'c')
    >>>
    >>> # make it favorable to (a AND b) OR (c AND d)
    >>> H.add_constraint_OR(['a', 'b'], ['c', 'd'])
    >>>
    >>> H
    {('b', 'a'): -2, (): 4, ('b',): -1, ('c',): -1,
     ('c', 'd'): -1, ('c', 'd', 'b', 'a'): 1}
    >>> Q = H.to_qubo()
    >>> Q
    {(): 4, (0,): -1, (2,): -1, (2, 3): 1, (4,): 6, (0, 4): -4,
     (1, 4): -4, (5,): 6, (2, 5): -4, (3, 5): -4, (4, 5): 1}
    >>> obj_value, sol = qubo_solver(Q)
    >>> sol
    {0: 1, 1: 1, 2: 1, 3: 0, 4: 1, 5: 0}
    >>> solution = H.convert_solution(sol)
    >>> solution
    {'b': 1, 'a': 1, 'c': 1, 'd': 0}

    """

    def __init__(self, *args, **kwargs):
        """__init__.

        This class deals with higher order binary optimization problems.
        Note that it is generally more efficient to initialize an empty HOBO
        object and then build the HOBO, rather than initialize a HOBO object
        with an already built dict.

        Parameters
        ----------
        arguments : define a dictionary with ``dict(*args, **kwargs)``.
            The dictionary will be initialized to follow all the convensions of
            the class. Alternatively, ``args[0]`` can be a HOBO object.

        Examples
        -------
        >>> hobo = HOBO()
        >>> hobo[('a',)] += 5
        >>> hobo[(0, 'a')] -= 2
        >>> hobo -= 1.5
        >>> hobo
        {('a',): 5, ('a', 0): -2, (): -1.5}
        >>> hobo.add_constraint_eq_zero({('a',): 1}, lam=5)
        >>> hobo
        {('a',): 10, ('a', 0): -2, (): -1.5}

        >>> hobo = HOBO({('a',): 5, (0, 'a', 1): -2, (): -1.5})
        >>> hobo
        {('a',): 5, ('a', 0, 1): -2, (): -1.5}

        """
        # use self.__class__ here because HOIO uses this code as well.
        super(self.__class__, self).__init__(*args, **kwargs)
        if len(args) == 1 and isinstance(args[0], self.__class__):
            self._constraints = args[0].constraints
            self._ancilla = args[0].num_ancillas
        else:
            self._ancilla, self._constraints = 0, {}

    def update(self, *args, **kwargs):
        """update.

        Update the HOBO but following all the conventions of this class.

        Parameters
        ----------
        arguments : defines a dictionary or HOBO.
            Ie ``d = dict(*args, **kwargs)``.
            Each element in d will be added in place to this instance following
            all the required convensions.

        """
        # use self.__class__ here because HOIO uses this code as well.
        super(self.__class__, self).update(*args, **kwargs)
        if len(args) == 1 and isinstance(args[0], self.__class__):
            for k, v in args[0]._constraints:
                self._constraints.setdefault(k, []).extend(v)

    @property
    def constraints(self):
        """constraints.

        Return the constraints of the HOBO.

        Return
        ------
        res : dict.
            The keys of ``res`` are some or all of
            ``'eq'``, ``'lt'``, ``'le'``, ``'gt'``, and ``'ge'``.
            The values are lists of ``qubovert.PUBO`` objects. For a
            given key, value pair ``k, v``, the ``v[i]`` element represents
            the PUBO ``v[i]`` being == 0 if ``k == 'eq'``,
            < 0 if ``k == 'lt'``, <= 0 if ``k == 'le'``,
            > 0 if ``k == 'gt'``, >= 0 if ``k == 'ge'``.

        """
        return {k: [x.copy() for x in v] for k, v in self._constraints.items()}

    def _append_constraint(self, key, constraint):
        """_append_constraint.

        Internal method to add a constraint to the constraints dictionary.

        Parameters
        ----------
        key : str.
            One of ``'eq'``, ``'lt'``, ``'le'``, ``'gt'``, or ``'ge'``.
        constraint : qubovert.PUBO object.

        """
        self._constraints.setdefault(key, []).append(constraint)

    def _pop_constraint(self, key):
        """_append_constraint.

        Internal method to remove the most recently added constraint.

        Parameters
        ----------
        key : str.
            One of ``'eq'``, ``'lt'``, ``'le'``, ``'gt'``, or ``'ge'``.

        """
        if self._constraints.get(key, []):
            self._constraints[key].pop()
            if not self._constraints[key]:
                self._constraints.pop(key)

    @property
    def num_ancillas(self):
        """num_ancillas.

        Return the number of ancilla variables introduced to the HOBO in
        order to enforce the inputted constraints.

        Returns
        -------
        num : int.
            Number of ancillas in the HOBO.

        """
        return self._ancilla

    @property
    def _next_ancilla(self):
        """_next_ancilla.

        Get the next available ancilla bit and increment.

        Return
        ------
        a : int.
            Ancilla bit label.

        """
        self._ancilla += 1
        return "__a%d" % (self._ancilla - 1)

    @classmethod
    def remove_ancilla_from_solution(cls, solution):
        """remove_ancilla_from_solution.

        Take a solution to the HOBO and remove all the ancilla variables, (
        represented by `_a` prefixes).

        Parameters
        ----------
        solution : dict.
            Must be the solution in terms of the original variables. Thus if
            ``solution`` is the solution to the ``self.to_pubo``,
            ``self.to_qubo``, ``self.to_hising``, or ``self.to_ising``
            formulations, then you should first call ``self.convert_solution``.
            See ``help(self.convert_solution)``.

        Return
        ------
        res : dict.
            The same as ``solution`` but with all the ancilla bits removed.

        """
        return {k: v for k, v in solution.items() if str(k)[:3] != "__a"}

    def is_solution_valid(self, solution):
        """is_solution_valid.

        Finds whether or not the given solution satisfies the constraints.

        Parameters
        ----------
        solution : dict.
            Must be the solution in terms of the original variables. Thus if
            ``solution`` is the solution to the ``self.to_pubo``,
            ``self.to_qubo``, ``self.to_hising``, or ``self.to_ising``
            formulations, then you should first call ``self.convert_solution``.
            See ``help(self.convert_solution)``.

        Return
        ------
        valid : bool.
            Whether or not the given solution satisfies the constraints.

        """
        if any(v.value(solution) != 0
               for v in self._constraints.get('eq', [])):
            return False

        if any(v.value(solution) >= 0
               for v in self._constraints.get("lt", [])):
            return False

        if any(v.value(solution) > 0
               for v in self._constraints.get("le", [])):
            return False

        if any(v.value(solution) <= 0
               for v in self._constraints.get("gt", [])):
            return False

        if any(v.value(solution) < 0
               for v in self._constraints.get("ge", [])):
            return False

        return True

    # override
    def __round__(self, ndigits=None):
        """round.

        Round values of the HOBO object.

        Parameters
        ----------
        ndigits : int.
            Number of decimal digits to round to.

        Returns
        -------
        res : HOBO object.
            Copy of self but with each value rounded to ``ndigits`` decimal
            digits. Each value has a type according to the docstring
            specifications of ``round``, see ``help(round)``.

        """
        # use self.__class__ here because HOIO uses this code as well.
        d = super(self.__class__, self).__round__(ndigits)
        d._constraints = self.constraints
        return d

    # override
    def subs(self, *args, **kwargs):
        """subs.

        Replace any ``sympy`` symbols that are used in the dict with values.
        Please see ``help(sympy.Symbol.subs)`` for more info.

        Parameters
        ----------
        arguments : substitutions.
            Same parameters as are inputted into ``sympy.Symbol.subs``.

        Returns
        -------
        res : HOBO object.
            Same as ``self`` but with all the symbols replaced with values.

        """
        # use self.__class__ here because HOIO uses this code as well.
        d = super(self.__class__, self).subs(*args, **kwargs)
        d._constraints = {
            k: [P.subs(*args, **kwargs) for P in v]
            for k, v in self._constraints.items()
        }
        return d

    def add_constraint_eq_zero(self,
                               P, lam=1,
                               bounds=None, suppress_warnings=False):
        r"""add_constraint_eq_zero.

        Enforce that ``P == 0`` by penalizing invalid solutions with ``lam``.

        Parameters
        ----------
        P : dict representing a PUBO.
            The PUBO constraint such that P == 0. Note that ``P`` will be
            converted to a ``qubovert.PUBO`` object if it is not already, thus
            it must follow the conventions, see ``help(qubovert.PUBO)``.
            Please note that if ``P`` contains any symbols, then ``bounds``
            must be supplied, since they cannot be determined when symbols
            are present.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the constraint.
        bounds : two element tuple (optional, defaults to None).
            A tuple ``(min, max)``, the minimum and maximum values that the
            PUBO ``P`` can take. If ``bounds`` is None, then they may be
            calculated (approximately).
        suppress_warnings : bool (optional, defaults to False).
            Whether or not to surpress warnings.

        Return
        ------
        self : HOBO.
            Updates the HOBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        The following enforces that :math:`\prod_{i=0}^{3} x_i == 0`.

        >>> H = HOBO()
        >>> H.add_constraint_eq_zero({(0, 1, 2, 3): 1})
        >>> H
        {(0, 1, 2, 3): 1}

        The following enforces that :math:`\sum_{i=1}^{3} i x_i x_{i+1} == 0`.

        >>> H = HOBO()
        >>> H.add_constraint_eq_zero({(1, 2): 1, (2, 3): 2, (3, 4): 3})
        >>> H
        {(1, 2): 1, (1, 2, 3): 4, (1, 2, 3, 4): 6,
         (2, 3): 4, (2, 3, 4): 12, (3, 4): 9}

        Here we show how operations can be strung together.

        >>> H = HOBO()
        >>> H.add_constraint_eq_zero(
                {(0, 1): 1}
            ).add_constraint_eq_zero(
                {(1, 2): 1, (): -1}
            )
        >>> H
        {(0, 1): 1, (1, 2): -1, (): 1}

        """
        P = PUBO(P)
        self._append_constraint("eq", P)

        min_val, max_val = _get_bounds(P, bounds)

        if min_val == max_val == 0:
            if not suppress_warnings:
                QUBOVertWarning.warn("Constraint is always satisfied")
        elif min_val > 0:
            if not suppress_warnings:
                QUBOVertWarning.warn("Constraint cannot be satisfied")
            self += lam * P
        elif max_val < 0:
            if not suppress_warnings:
                QUBOVertWarning.warn("Constraint cannot be satisfied")
            self -= lam * P
        elif min_val == 0:
            self += lam * P
        elif max_val == 0:
            self -= lam * P
        else:
            self += lam * P ** 2

        return self

    def add_constraint_lt_zero(self,
                               P, lam=1, log_trick=True,
                               bounds=None, suppress_warnings=False):
        r"""add_constraint_lt_zero.

        Enforce that ``P < 0`` by penalizing invalid solutions with ``lam``.
        See Notes below for more details.

        Parameters
        ----------
        P : dict representing a PUBO.
            The PUBO constraint such that P < 0. Note that ``P`` will be
            converted to a ``qubovert.PUBO`` object if it is not already, thus
            it must follow the conventions, see ``help(qubovert.PUBO)``.
            Please note that if ``P`` contains any symbols, then ``bounds``
            must be supplied, since they cannot be determined when symbols
            are present.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the constraint.
        log_trick : bool (optional, defaults to True).
            Whether or not to use the log trick to enforce the inequality
            constraint. See Notes below for more details.
        bounds : two element tuple (optional, defaults to None).
            A tuple ``(min, max)``, the minimum and maximum values that the
            PUBO ``P`` can take. If ``bounds`` is None, then they will be
            calculated (approximately), or if either of the elements of
            ``bounds`` is None, then that element will be calculated
            (approximately).
        suppress_warnings : bool (optional, defaults to False).
            Whether or not to surpress warnings.

        Return
        ------
        self : HOBO.
            Updates the HOBO in place, but returns ``self`` so that operations
            can be strung together.

        Notes
        -----
        - There is no general way to enforce non integer inequality
          constraints. Thus this function is only guarenteed to work for
          integer inequality constraints (ie constraints of the form
          :math:`x_0 + 2x_1 + ... < 0`). However, it can be used for non
          integer inequality constraints, but it is recommended that the value
          of ``lam`` be set small, since valid solutions may still recieve a
          penalty to the objective function. For example,

          >>> H = HOBO()
          >>> H.add_constraint_lt_zero({(0,): 1, (1,): 2, (2,): -.5, (): .4})
          >>> test_sol = {0: 0, 1: 0, 2: 1}
          >>> H.is_solution_valid(test_sol)
          True
          >>> H.value(test_sol)
          0.01

          {0: 0, 1: 0, 2: 1} is a valid solution to ``H``, but it will still
          cause a nonzero penalty to be added to the objective function.

        - To enforce the inequality constraint, ancilla bits will be
          introduced (labels with `_a`). If ``log_trick`` is ``True``, then
          approximately :math:`\log_2 |\min_x \text{P.value(x)}|`
          ancilla bits will be used. If ``log_trick`` is ``False``, then
          approximately :math:`|\min_x \text{P.value(x)}|` ancilla
          bits will be used.

        Examples
        --------
        Enforce that :math:`-x_a x_b x_c + x_a -4x_a x_b + 3x_c < 2`.

        >>> H = HOBO().add_constraint_lt_zero(
                {('a', 'b', 'c'): -1, ('a',): 1,
                 ('a', 'b'): -4, ('c',): 3, (): -2}
            )
        >>> H
        {('b', 'c', 'a'): -19, ('b', '__a0', 'c', 'a'): -2,
         ('b', 'c', 'a', '__a1'): -4, ('a',): -3, ('b', 'a'): 24,
         ('c', 'a'): 6, ('__a0', 'a'): 2, ('a', '__a1'): 4,
         ('b', '__a0', 'a'): -8, ('b', 'a', '__a1'): -16, ('c',): -3,
         ('__a0', 'c'): 6, ('c', '__a1'): 12, (): 4, ('__a0',): -3,
         ('__a1',): -4, ('__a0', '__a1'): 4}
        >>> H.is_solution_valid({'b': 0, 'c': 0, 'a': 1})
        True
        >>> H.is_solution_valid({'b': 0, 'c': 1, 'a': 1})
        False

        """
        P = PUBO(P)
        self._append_constraint("lt", P)

        min_val, max_val = _get_bounds(P, bounds)

        if min_val >= 0:
            if not suppress_warnings:
                QUBOVertWarning.warn("Constraint cannot be satisfied")
            self += lam * P
        elif max_val < 0:
            if not suppress_warnings:
                QUBOVertWarning.warn("Constraint is always satisfied")
        else:
            if int(min_val) == min_val:
                # copy P, don't do +=
                P = P + 1
                min_val += 1
                max_val += 1
            self.add_constraint_le_zero(
                P, lam=lam, log_trick=log_trick,
                bounds=(min_val, max_val), suppress_warnings=True
            )
            self._pop_constraint("le")

        return self

    def add_constraint_le_zero(self,
                               P, lam=1, log_trick=True,
                               bounds=None, suppress_warnings=False):
        r"""add_constraint_le_zero.

        Enforce that ``P <= 0`` by penalizing invalid solutions with ``lam``.

        Parameters
        ----------
        P : dict representing a PUBO.
            The PUBO constraint such that P <= 0. Note that ``P`` will be
            converted to a ``qubovert.PUBO`` object if it is not already, thus
            it must follow the conventions, see ``help(qubovert.PUBO)``.
            Please note that if ``P`` contains any symbols, then ``bounds``
            must be supplied, since they cannot be determined when symbols
            are present.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the constraint.
        log_trick : bool (optional, defaults to True).
            Whether or not to use the log trick to enforce the inequality
            constraint. See Notes below for more details.
        bounds : two element tuple (optional, defaults to None).
            A tuple ``(min, max)``, the minimum and maximum values that the
            PUBO ``P`` can take. If ``bounds`` is None, then they will be
            calculated (approximately), or if either of the elements of
            ``bounds`` is None, then that element will be calculated
            (approximately).
        suppress_warnings : bool (optional, defaults to False).
            Whether or not to surpress warnings.

        Return
        ------
        self : HOBO.
            Updates the HOBO in place, but returns ``self`` so that operations
            can be strung together.

        Notes
        -----
        - There is no general way to enforce non integer inequality
          constraints. Thus this function is only guarenteed to work for
          integer inequality constraints (ie constraints of the form
          :math:`x_0 + 2x_1 + ... \leq 0`). However, it can be used for non
          integer inequality constraints, but it is recommended that the value
          of ``lam`` be set small, since valid solutions may still recieve a
          penalty to the objective function. For example,

          >>> H = HOBO()
          >>> H.add_constraint_le_zero({(0,): 1, (1,): 2, (2,): -1.5, (): .4})
          >>> H
          {(0,): 1.7999999999999998, (0, 1): 4, (0, 2): -3.0, (0, '__a0'): 2,
           (1,): 5.6, (1, 2): -6.0, (1, '__a0'): 4, (2,): 1.0499999999999998,
           (2, '__a0'): -3.0, (): 0.16000000000000003, ('__a0',): 1.8}
          >>> test_sol = {0: 0, 1: 0, 2: 1, '__a0': 1}
          >>> H.is_solution_valid(test_sol)
          True
          >>> H.value(test_sol)
          0.01

          {0: 0, 1: 0, 2: 1} is a valid solution to ``H``, but it will still
          cause a nonzero penalty to be added to the objective function.

        - To enforce the inequality constraint, ancilla bits will be
          introduced (labels with `_a`). If ``log_trick`` is ``True``, then
          approximately :math:`\log_2 |\min_x \text{P.value(x)}|`
          ancilla bits will be used. If ``log_trick`` is ``False``, then
          approximately :math:`|\min_x \text{P.value(x)}|` ancilla
          bits will be used.

        Examples
        --------
        Enforce that :math:`-x_a x_b x_c + x_a -4x_a x_b + 3x_c \leq 2`.

        >>> H = HOBO().add_constraint_le_zero(
                {('a', 'b', 'c'): -1, ('a',): 1,
                 ('a', 'b'): -4, ('c',): 3, (): -2}
            )
        >>> H
        {('b', 'c', 'a'): -19, ('b', 'c', 'a', '__a0'): -2,
         ('__a1', 'b', 'c', 'a'): -4, ('__a2', 'b', 'c', 'a'): -8, ('a',): -3,
         ('b', 'a'): 24, ('c', 'a'): 6, ('a', '__a0'): 2, ('__a1', 'a'): 4,
         ('__a2', 'a'): 8, ('b', 'a', '__a0'): -8, ('__a1', 'b', 'a'): -16,
         ('__a2', 'b', 'a'): -32, ('c',): -3, ('c', '__a0'): 6,
         ('__a1', 'c'): 12, ('__a2', 'c'): 24, (): 4, ('__a0',): -3,
         ('__a1',): -4, ('__a1', '__a0'): 4, ('__a2', '__a0'): 8,
         ('__a2', '__a1'): 16}
        >>> H.is_solution_valid({'b': 0, 'c': 0, 'a': 1})
        True
        >>> H.is_solution_valid({'b': 0, 'c': 1, 'a': 1})
        False

        """
        P = PUBO(P)
        self._append_constraint("le", P)

        if _special_constraints_le_zero(self, P, lam):
            return self

        min_val, max_val = _get_bounds(P, bounds)

        if min_val > 0:
            if not suppress_warnings:
                QUBOVertWarning.warn("Constraint cannot be satisfied")
            self += lam * P
        elif max_val <= 0:
            if not suppress_warnings:
                QUBOVertWarning.warn("Constraint is always satisfied")
        else:
            # don't mutate the P that we put in self._constraints
            P = P.copy()
            if log_trick and min_val:
                for i in range(int(ceil(log2(-min_val + 1)))):
                    P[(self._next_ancilla,)] += pow(2, i)
                    max_val += pow(2, i)
            elif min_val:
                for _ in range(int(ceil(-min_val))):
                    P[(self._next_ancilla,)] += 1
                    max_val += 1

            self.add_constraint_eq_zero(
                P, lam=lam,
                bounds=(min_val, max_val), suppress_warnings=True
            )
            self._pop_constraint("eq")

        return self

    def add_constraint_gt_zero(self,
                               P, lam=1, log_trick=True,
                               bounds=None, suppress_warnings=False):
        r"""add_constraint_gt_zero.

        Enforce that ``P > 0`` by penalizing invalid solutions with ``lam``.

        Parameters
        ----------
        P : dict representing a PUBO.
            The PUBO constraint such that P > 0. Note that ``P`` will be
            converted to a ``qubovert.PUBO`` object if it is not already, thus
            it must follow the conventions, see ``help(qubovert.PUBO)``.
            Please note that if ``P`` contains any symbols, then ``bounds``
            must be supplied, since they cannot be determined when symbols
            are present.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the constraint.
        log_trick : bool (optional, defaults to True).
            Whether or not to use the log trick to enforce the inequality
            constraint. See Notes below for more details.
        bounds : two element tuple (optional, defaults to None).
            A tuple ``(min, max)``, the minimum and maximum values that the
            PUBO ``P`` can take. If ``bounds`` is None, then they will be
            calculated (approximately), or if either of the elements of
            ``bounds`` is None, then that element will be calculated
            (approximately).
        suppress_warnings : bool (optional, defaults to False).
            Whether or not to surpress warnings.

        Return
        ------
        self : HOBO.
            Updates the HOBO in place, but returns ``self`` so that operations
            can be strung together.

        Notes
        -----
        - There is no general way to enforce non integer inequality
          constraints. Thus this function is only guarenteed to work for
          integer inequality constraints (ie constraints of the form
          :math:`x_0 + 2x_1 + ... > 0`). However, it can be used for non
          integer inequality constraints, but it is recommended that the value
          of ``lam`` be set small, since valid solutions may still recieve a
          penalty to the objective function. For example,

          >>> H = HOBO()
          >>> H.add_constraint_gt_zero({(0,): -1, (1,): -2, (2,): .5, (): -.4})
          >>> test_sol = {0: 0, 1: 0, 2: 1}
          >>> H.is_solution_valid(test_sol)
          True
          >>> H.value(test_sol)
          0.01

          {0: 0, 1: 0, 2: 1} is a valid solution to ``H``, but it will still
          cause a nonzero penalty to be added to the objective function.

        - To enforce the inequality constraint, ancilla bits will be
          introduced (labels with `_a`). If ``log_trick`` is ``True``, then
          approximately :math:`\log_2 |\max_x \text{P.value(x)}|`
          ancilla bits will be used. If ``log_trick`` is ``False``, then
          approximately :math:`|\max_x \text{P.value(x)}|` ancilla
          bits will be used.

        Examples
        --------
        Enforce that :math:`x_a x_b x_c - x_a + 4x_a x_b - 3x_c > -2`.

        >>> H = HOBO().add_constraint_gt_zero(
                {('a', 'b', 'c'): 1, ('a',): -1,
                 ('a', 'b'): 4, ('c',): -3, (): 2}
            )
        >>> H
        {('b', 'c', 'a'): -19, ('b', '__a0', 'c', 'a'): -2,
         ('b', 'c', 'a', '__a1'): -4, ('a',): -3, ('b', 'a'): 24,
         ('c', 'a'): 6, ('__a0', 'a'): 2, ('a', '__a1'): 4,
         ('b', '__a0', 'a'): -8, ('b', 'a', '__a1'): -16, ('c',): -3,
         ('__a0', 'c'): 6, ('c', '__a1'): 12, (): 4, ('__a0',): -3,
         ('__a1',): -4, ('__a0', '__a1'): 4}
        >>> H.is_solution_valid({'b': 0, 'c': 0, 'a': 1})
        True
        >>> H.is_solution_valid({'b': 0, 'c': 1, 'a': 1})
        False

        """
        P = PUBO(P)
        self._append_constraint("gt", P)
        min_val, max_val = _get_bounds(P, bounds)
        bounds = -max_val, -min_val
        self.add_constraint_lt_zero(
            -P, lam=lam, log_trick=log_trick,
            bounds=bounds, suppress_warnings=suppress_warnings
        )
        self._pop_constraint("lt")
        return self

    def add_constraint_ge_zero(self,
                               P, lam=1, log_trick=True,
                               bounds=None, suppress_warnings=False):
        r"""add_constraint_ge_zero.

        Enforce that ``P >= 0`` by penalizing invalid solutions with ``lam``.

        Parameters
        ----------
        P : dict representing a PUBO.
            The PUBO constraint such that P >= 0. Note that ``P`` will be
            converted to a ``qubovert.PUBO`` object if it is not already, thus
            it must follow the conventions, see ``help(qubovert.PUBO)``.
            Please note that if ``P`` contains any symbols, then ``bounds``
            must be supplied, since they cannot be determined when symbols
            are present.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the constraint.
        log_trick : bool (optional, defaults to True).
            Whether or not to use the log trick to enforce the inequality
            constraint. See Notes below for more details.
        bounds : two element tuple (optional, defaults to None).
            A tuple ``(min, max)``, the minimum and maximum values that the
            PUBO ``P`` can take. If ``bounds`` is None, then they will be
            calculated (approximately), or if either of the elements of
            ``bounds`` is None, then that element will be calculated
            (approximately), or if either of the elements of
            ``bounds`` is None, then that element will be calculated
            (approximately).
        suppress_warnings : bool (optional, defaults to False).
            Whether or not to surpress warnings.

        Return
        ------
        self : HOBO.
            Updates the HOBO in place, but returns ``self`` so that operations
            can be strung together.

        Notes
        -----
        - There is no general way to enforce non integer inequality
          constraints. Thus this function is only guarenteed to work for
          integer inequality constraints (ie constraints of the form
          :math:`x_0 + 2x_1 + ... \geq 0`). However, it can be used for non
          integer inequality constraints, but it is recommended that the value
          of ``lam`` be set small, since valid solutions may still recieve a
          penalty to the objective function. For example,

          >>> H = HOBO()
          >>> H.add_constraint_ge_zero({(0,): -1, (1,): -2, (2,):1.5, (): -.4})
          >>> H
          {(0,): 1.7999999999999998, (0, 1): 4, (0, 2): -3.0, (0, '__a0'): 2,
           (1,): 5.6, (1, 2): -6.0, (1, '__a0'): 4, (2,): 1.0499999999999998,
           (2, '__a0'): -3.0, (): 0.16000000000000003, ('__a0',): 1.8}
          >>> test_sol = {0: 0, 1: 0, 2: 1, '__a0': 1}
          >>> H.is_solution_valid(test_sol)
          True
          >>> H.value(test_sol)
          0.01

          {0: 0, 1: 0, 2: 1} is a valid solution to ``H``, but it will still
          cause a nonzero penalty to be added to the objective function.

        - To enforce the inequality constraint, ancilla bits will be
          introduced (labels with `_a`). If ``log_trick`` is ``True``, then
          approximately :math:`\log_2 |\max_x \text{P.value(x)}|`
          ancilla bits will be used. If ``log_trick`` is ``False``, then
          approximately :math:`|\max_x \text{P.value(x)}|` ancilla
          bits will be used.

        Examples
        --------
        Enforce that :math:`x_a x_b x_c - x_a + 4x_a x_b - 3x_c \geq -2`.

        >>> H = HOBO().add_constraint_ge_zero(
                {('a', 'b', 'c'): 1, ('a',): -1,
                 ('a', 'b'): 4, ('c',): -3, (): 2}
            )
        >>> H
        {('b', 'c', 'a'): -19, ('b', 'c', 'a', '__a0'): -2,
         ('__a1', 'b', 'c', 'a'): -4, ('__a2', 'b', 'c', 'a'): -8, ('a',): -3,
         ('b', 'a'): 24, ('c', 'a'): 6, ('a', '__a0'): 2, ('__a1', 'a'): 4,
         ('__a2', 'a'): 8, ('b', 'a', '__a0'): -8, ('__a1', 'b', 'a'): -16,
         ('__a2', 'b', 'a'): -32, ('c',): -3, ('c', '__a0'): 6,
         ('__a1', 'c'): 12, ('__a2', 'c'): 24, (): 4, ('__a0',): -3,
         ('__a1',): -4, ('__a1', '__a0'): 4, ('__a2', '__a0'): 8,
         ('__a2', '__a1'): 16}
        >>> H.is_solution_valid({'b': 0, 'c': 0, 'a': 1})
        True
        >>> H.is_solution_valid({'b': 0, 'c': 1, 'a': 1})
        False

        """
        P = PUBO(P)
        self._append_constraint("ge", P)
        min_val, max_val = _get_bounds(P, bounds)
        bounds = -max_val, -min_val
        self.add_constraint_le_zero(
            -P, lam=lam, log_trick=log_trick,
            bounds=bounds, suppress_warnings=suppress_warnings
        )
        self._pop_constraint("le")
        return self

    def add_constraint_eq_AND(self, a, *variables, lam=1):
        r"""add_constraint_eq_AND.

        Add a penalty to the HOBO that enforces that
        :math:`a = v_0 \land v_1 \land v_2 \land ...`,
        with a penalty factor
        ``lam``, where ``v_1 = variables[0]``, ``v_2 = variables[1]``, ...

        Parameters
        ----------
        a : any hashable object or a list of hashable objects.
            The label for binary variables ``a``.
        *variables : arguments.
            Each element of variables is a hashable object or a list of
            hashable objects. They are the label of the binary variables.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : HOBO.
            Updates the HOBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = HOBO()
        >>> H.add_constraint_eq_AND('a', 'b', 'c')  # enforce a == b AND c

        >>> H = HOBO()
        >>> # enforce (a AND b AND c AND d) == 'e'
        >>> H.add_constraint_eq_AND('e', ['a', 'b'], ['c', 'd'])

        >>> H = HOBO()
        >>> # enforce (a AND b AND c AND d) == 'e'
        >>> H.add_constraint_eq_AND('e', 'a', 'b', 'c', 'd')

        >>> H = HOBO()
        >>> # enforce (a AND b AND c AND d) == 'e' AND 'f'
        >>> H.add_constraint_eq_AND(['e', 'f'], ['a', 'b'], ['c', 'd'])

        References
        ----------
        https://arxiv.org/pdf/1307.8041.pdf equation 6.

        """
        n = len(variables)
        if n < 2:
            raise ValueError("Must supply at least two variables to AND. "
                             "See ``add_constraint_eq_ONE`` for less.")
        a = _create_tuple(a)
        b, c = (), ()
        for v in variables[:n // 2]:
            b += _create_tuple(v)
        for v in variables[n // 2:]:
            c += _create_tuple(v)

        P = PUBO({a: 3, b+c: 1, b+a: -2, c+a: -2})

        return self.add_constraint_eq_zero(P, lam, bounds=(0, 3))

    def add_constraint_eq_OR(self, a, *variables, lam=1):
        r"""add_constraint_eq_OR.

        Add a penalty to the HOBO that enforces that
        :math:`v_0 \lor v_1 \lor v_2 \lor ... == a`,
        with a penalty factor ``lam``, where ``v_0 = variables[0]``,
        ``v_1 = variables[1]``, ...

        Parameters
        ----------
        a : any hashable object or a list of hashable objects.
            The label for binary variables ``a``.
        *variables : arguments.
            Each element of variables is a hashable object or a list of
            hashable objects. They are the label of the binary variables.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : HOBO.
            Updates the HOBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = HOBO()
        >>> H.add_constraint_eq_OR('a', 'b', 'c')  # enforce a == b OR c

        >>> H = HOBO()
        >>> # enforce (a AND b) OR (c AND d) == 'e'
        >>> H.add_constraint_eq_OR('e', ['a', 'b'], ['c', 'd'])

        >>> H = HOBO()
        >>> # enforce (a AND b) OR (c AND d) OR e == f
        >>> H.add_constraint_eq_OR('f', ['a', 'b'], ['c', 'd'], 'e')

        >>> H = HOBO()
        >>> # enforce (a AND b) OR (c AND d) == ('e' AND 'f')
        >>> H.add_constraint_eq_OR(['e', 'f'], ['a', 'b'], ['c', 'd'])

        """
        n = len(variables)
        if n < 2:
            raise ValueError("Must supply at least two variables to OR.")

        a = _create_tuple(a)
        if n == 2:
            b, c = tuple(_create_tuple(x) for x in variables)
            P = PUBO({a: 1, b: 1, c: 1, b+c: 1, b+a: -2, c+a: -2})
            bounds = 0, 3
        else:
            P = HOBO().add_constraint_NOR(*variables) - {a: 1}
            bounds = -1, 1

        return self.add_constraint_eq_zero(P, lam=lam, bounds=bounds)

    def add_constraint_eq_XOR(self, a, *variables, lam=1):
        r"""add_constraint_eq_XOR.

        Add a penalty to the HOBO that enforces that
        :math:`v_0 \oplus v_1 \oplus ... == a`
        with a penalty factor ``lam``, where ``v_0 = variables[0]``,
        ``v_1 = variables[1]``, ...

        Parameters
        ----------
        a : any hashable object or a list of hashable objects.
            The label for binary variables ``a``.
        *variables : arguments.
            Each element of variables is a hashable object or a list of
            hashable objects. They are the label of the binary variables.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : HOBO.
            Updates the HOBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = HOBO()
        >>> H.add_constraint_eq_XOR('a', 'b', 'c')  # enforce a == b XOR c

        >>> H = HOBO()
        >>> # enforce (a AND b) XOR (c AND d) == 'e'
        >>> H.add_constraint_eq_XOR('e', ['a', 'b'], ['c', 'd'])

        >>> H = HOBO()
        >>> # enforce (a AND b) XOR (c AND d) == ('e' AND 'f')
        >>> H.add_constraint_eq_XOR(['e', 'f'], ['a', 'b'], ['c', 'd'])

        """
        P = HOBO().add_constraint_XNOR(*variables) - {_create_tuple(a): 1}
        return self.add_constraint_eq_zero(P, lam=lam, bounds=(-1, 1))

    def add_constraint_eq_ONE(self, a, b, lam=1):
        r"""add_constraint_eq_ONE.

        Add a penalty to the HOBO that enforces that :math:`a == b` with
        a penalty factor ``lam``.

        Parameters
        ----------
        a : any hashable object or a list of hashable objects.
            The label for binary variables ``a``.
        b : any hashable object or a list of hashable objects.
            The label for binary variables ``b``.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : HOBO.
            Updates the HOBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = HOBO()
        >>> H.add_constraint_eq_ONE('a', 'b')  # enforce a == b

        >>> H = HOBO()
        >>> H.add_constraint_eq_ONE(['a', 'b'], 'c')  # enforce (a AND b) == c

        >>> H = HOBO()
        >>> # enforce (a AND b) == (c AND d)
        >>> H.add_constraint_eq_ONE(['a', 'b'], ['c', 'd'])

        """
        P = PUBO({_create_tuple(a): 1, _create_tuple(b): -1})
        return self.add_constraint_eq_zero(P, lam=lam, bounds=(-1, 1))

    def add_constraint_eq_NAND(self, a, *variables, lam=1):
        r"""add_constraint_eq_NAND.

        Add a penalty to the HOBO that enforces that
        :math:`\lnot (v_0 \land v_1 \land v_2 \land ...) == a`, with a
        penalty factor ``lam``, where ``v_0 = variables[0]``,
        ``v_1 = variables[1]``, ...

        Parameters
        ----------
        a : any hashable object or a list of hashable objects.
            The label for binary variables ``a``.
        *variables : arguments.
            Each element of variables is a hashable object or a list of
            hashable objects. They are the label of the binary variables.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : HOBO.
            Updates the HOBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = HOBO()
        >>> H.add_constraint_eq_NAND('a', 'b', 'c')  # enforce a == b NAND c

        >>> H = HOBO()
        >>> # enforce (a AND b) NAND (c AND d) == 'e'
        >>> H.add_constraint_eq_NAND('e', ['a', 'b'], ['c', 'd'])

        >>> H = HOBO()
        >>> # enforce (a AND b) NAND (c AND d) NAND e == f
        >>> H.add_constraint_eq_NAND('f', ['a', 'b'], ['c', 'd'], 'e')

        >>> H = HOBO()
        >>> # enforce (a AND b) NAND (c AND d) == 'e' AND 'f'
        >>> H.add_constraint_eq_NAND(['e', 'f'], ['a', 'b'], ['c', 'd'])

        """
        n = len(variables)
        if n < 2:
            raise ValueError("Must supply at least two variables to NAND. "
                             "See ``add_constraint_eq_NOT`` for less.")
        a = _create_tuple(a)
        b, c = (), ()
        for v in variables[:n // 2]:
            b += _create_tuple(v)
        for v in variables[n // 2:]:
            c += _create_tuple(v)

        P = PUBO({(): 3, b: -2, c: -2, a: -3, b+c: 1, b+a: 2, c+a: 2})
        return self.add_constraint_eq_zero(P, lam=lam, bounds=(0, 3))

    def add_constraint_eq_NOR(self, a, *variables, lam=1):
        r"""add_constraint_eq_NOR.

        Add a penalty to the HOBO that enforces that
        :math:`\lnot(v_0 \lor v_1 \lor v_2 \lor ...) == a`,
        with a penalty factor ``lam``, where ``v_0 = variables[0]``,
        ``v_1 = variables[1]``, ...

        Parameters
        ----------
        a : any hashable object or a list of hashable objects.
            The label for binary variables ``a``.
        *variables : arguments.
            Each element of variables is a hashable object or a list of
            hashable objects. They are the label of the binary variables.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : HOBO.
            Updates the HOBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = HOBO()
        >>> H.add_constraint_eq_NOR('a', 'b', 'c')  # enforce a == b NOR c

        >>> H = HOBO()
        >>> # enforce (a AND b) NOR (c AND d) == 'e'
        >>> H.add_constraint_eq_NOR('e', ['a', 'b'], ['c', 'd'])

        >>> H = HOBO()
        >>> # enforce (a AND b) NOR (c AND d) NOR e == f
        >>> H.add_constraint_eq_NOR('f', ['a', 'b'], ['c', 'd'], 'e')

        >>> H = HOBO()
        >>> # enforce (a AND b) NOR (c AND d) == ('e' AND 'f')
        >>> H.add_constraint_eq_NOR(['e', 'f'], ['a', 'b'], ['c', 'd'])

        """
        n = len(variables)
        if n < 2:
            raise ValueError("Must supply at least two variables to NOR.")

        a = _create_tuple(a)
        if n == 2:
            b, c = tuple(_create_tuple(x) for x in variables)
            P = PUBO({a: -1, b: -1, c: -1, (): 1, b+c: 1, b+a: 2, c+a: 2})
            bounds = 0, 3
        else:
            P = HOBO().add_constraint_OR(*variables) - {a: 1}
            bounds = -1, 1

        return self.add_constraint_eq_zero(P, lam=lam, bounds=bounds)

    def add_constraint_eq_XNOR(self, a, *variables, lam=1):
        r"""add_constraint_eq_XNOR.

        Add a penalty to the HOBO that enforces that
        :math:\lnot(`v_0 \oplus v_1 \oplus ...) == a`
        with a penalty factor ``lam``, where ``v_0 = variables[0]``,
        ``v_1 = variables[1]``, ...

        Parameters
        ----------
        a : any hashable object or a list of hashable objects.
            The label for binary variables ``a``.
        *variables : arguments.
            Each element of variables is a hashable object or a list of
            hashable objects. They are the label of the binary variables.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : HOBO.
            Updates the HOBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = HOBO()
        >>> H.add_constraint_eq_XNOR('a', 'b', 'c')  # enforce a == b XNOR c

        >>> H = HOBO()
        >>> # enforce (a AND b) XNOR (c AND d) == 'e'
        >>> H.add_constraint_eq_XNOR('e', ['a', 'b'], ['c', 'd'])

        >>> H = HOBO()
        >>> # enforce (a AND b) XNOR (c AND d) == ('e' AND 'f')
        >>> H.add_constraint_eq_XNOR(['e', 'f'], ['a', 'b'], ['c', 'd'])

        """
        P = HOBO().add_constraint_XOR(*variables) - {_create_tuple(a): 1}
        return self.add_constraint_eq_zero(P, lam=lam, bounds=(-1, 1))

    def add_constraint_eq_NOT(self, a, b, lam=1):
        r"""NOT.

        Add a penalty to the HOBO that enforces that :math:`\lnot a == b` with
        a penalty factor ``lam``.

        Parameters
        ----------
        a : any hashable object or a list of hashable objects.
            The label for binary variables ``a``.
        b : any hashable object or a list of hashable objects.
            The label for binary variables ``b``.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : HOBO.
            Updates the HOBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = HOBO()
        >>> H.add_constraint_eq_NOT('a', 'b')  # enforce NOT(a) == b

        >>> H = HOBO()
        >>> # enforce NOT (a AND b) == c
        >>> H.add_constraint_eq_NOT(['a', 'b'], 'c')

        >>> H = HOBO()
        >>> # enforce NOT(a AND b) == (c AND d)
        >>> H.add_constraint_eq_NOT(['a', 'b'], ['c', 'd'])

        """
        P = HOBO().add_constraint_ONE(a) - {_create_tuple(b): 1}
        return self.add_constraint_eq_zero(P, lam=lam, bounds=(-1, 1))

    def add_constraint_AND(self, *variables, lam=1):
        r"""add_constraint_AND.

        Add a penalty to the HOBO that is only zero when
        :math:`a \land b \land c \land ...` is True, with a penalty factor
        ``lam``, where ``a = variables[0]``, ``b = variables[1]``, etc.

        Parameters
        ----------
        *variables : arguments.
            Each element of variables is a hashable object or a list of
            hashable objects. They are the label of the binary variables.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : HOBO.
            Updates the HOBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = HOBO()
        >>> H.add_constraint_AND('a', 'b')  # enforce a AND b
        >>> H
        {('b', 'a'): -1, (): 1}

        >>> H = HOBO()
        >>> H.add_constraint_AND(['a', 'b'], ['c', 'd'])
        >>> # enforce a AND b AND c AND d
        >>> H
        {('c', 'd', 'b', 'a'): -1, (): 1}

        >>> H = HOBO()
        >>> H.add_constraint_AND('a', 'b', 'c', 'd')
        >>> # enforce a AND b AND c AND d

        """
        t = ()
        for x in variables:
            t += _create_tuple(x)
        return self.add_constraint_ONE(list(t), lam=lam)

    def add_constraint_OR(self, *variables, lam=1):
        r"""add_constraint_OR.

        Add a penalty to the HOBO that is only nonzero when
        :math:`a \lor b \lor c \lor d \lor ...` is True, with a penalty factor
        ``lam``, where ``a = variables[0]``, ``b = variables[1]``, etc.

        Parameters
        ----------
        *variables : arguments.
            Each element of variables is a hashable object or a list of
            hashable objects. They are the label of the binary variables.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : HOBO.
            Updates the HOBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = HOBO()
        >>> H.add_constraint_OR('a', 'b')  # enforce a OR b
        >>> H
        {('a',): -1, ('b',): -1, ('b', 'a'): 1, (): 1}

        >>> H = HOBO()
        >>> H.add_constraint_OR(['a', 'b'], ['c', 'd'])
        >>> # enforce (a AND b) OR (c AND d)
        >>> H
        {('b', 'a'): -1, ('c', 'd'): -1, ('c', 'd', 'b', 'a'): 1, (): 1}

        >>> H = HOBO()
        >>> H.add_constraint_OR('a', 'b', 'c', 'd')  # enforce a OR b OR c OR d

        >>> H = HOBO()
        >>> # enforce (a AND b) OR (c AND d) OR (e AND f AND g)
        >>> H.add_constraint_OR(['a', 'b'], ['c', 'd'], ['e', 'f', 'g'])

        """
        variables = tuple({_create_tuple(x): 1} for x in variables)
        P = 1 - OR(*variables)
        return self.add_constraint_eq_zero(P, lam=lam, bounds=(0, 1))

    def add_constraint_XOR(self, *variables, lam=1):
        r"""add_constraint_XOR.

        Add a penalty to the HOBO that is only nonzero when
        :math:`v_0 \oplus v_1 \oplus ... \oplus v_n` is True, with a penalty
        factor ``lam``, where ``v_0 = variables[0]``, ``v_1 = variables[1]``,
        ..., ``v_n = variables[-1]``. See ``qubovert.sat.XOR`` for the
        XOR convention that qubovert uses for more than two inputs.

        Parameters
        ----------
        *variables : arguments.
            Each element of variables is a hashable object or a list of
            hashable objects. They are the label of the binary variables.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : HOBO.
            Updates the HOBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = HOBO()
        >>> H.add_constraint_XOR('a', 'b')  # enforce a XOR b

        >>> H = HOBO()
        >>> H.add_constraint_XOR(['a', 'b'], ['c', 'd'])
        >>> # enforce (a AND b) XOR (c AND d)

        >>> H = HOBO()
        >>> H.add_constraint_XOR('a', 'b', 'c')  # enforce a XOR b XOR c

        """
        variables = tuple({_create_tuple(x): 1} for x in variables)
        P = 1 - XOR(*variables)
        return self.add_constraint_eq_zero(P, lam=lam, bounds=(0, 1))

    def add_constraint_ONE(self, a, lam=1):
        r"""add_constraint_ONE.

        Add a penalty to the HOBO that is only nonzero when :math:`a == 1` is
        True, with a penalty factor ``lam``.

        Parameters
        ----------
        a : any hashable object or a list of hashable objects.
            The label for binary variables ``a``.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : HOBO.
            Updates the HOBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = HOBO()
        >>> H.add_constraint_ONE('a')  # enforce a

        >>> H = HOBO()
        >>> H.add_constraint_ONE(['a', 'b'])  # enforce (a AND b)

        """
        P = PUBO({(): 1, _create_tuple(a): -1})
        return self.add_constraint_eq_zero(P, lam=lam, bounds=(0, 1))

    def add_constraint_NAND(self, *variables, lam=1):
        r"""add_constraint_NAND.

        Add a penalty to the HOBO that is only zero when
        :math:`\lnot (a \land b \land c \land ...)` is True, with a penalty
        factor ``lam``, where ``a = variables[0]``, ``b = variables[1]``, etc.

        Parameters
        ----------
        *variables : arguments.
            Each element of variables is a hashable object or a list of
            hashable objects. They are the label of the binary variables.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : HOBO.
            Updates the HOBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = HOBO()
        >>> H.add_constraint_NAND('a', 'b')  # enforce a NAND b

        >>> H = HOBO()
        >>> H.add_constraint_NAND(['a', 'b'], ['c', 'd'])
        >>> # enforce (a AND b) NAND (c AND d)

        >>> H = HOBO()
        >>> H.add_constraint_NAND('a', 'b', 'c', 'd')
        >>> # enforce a NAND b NAND c NAND d

        >>> H = HOBO()
        >>> # enforce (a AND b) NAND (c AND d AND e) NAND f
        >>> H.add_constraint_NAND(['a', 'b'], ['c', 'd', 'e'], 'f')

        """
        t = ()
        for x in variables:
            t += _create_tuple(x)
        return self.add_constraint_NOT(list(t), lam=lam)

    def add_constraint_NOR(self, *variables, lam=1):
        r"""add_constraint_NOR.

        Add a penalty to the HOBO that is only nonzero when
        :math:`\lnot(a \lor b \lor c \lor d \lor ...)` is True, with a penalty
        factor ``lam``, where ``a = variables[0]``, ``b = variables[1]``, etc.

        Parameters
        ----------
        *variables : arguments.
            Each element of variables is a hashable object or a list of
            hashable objects. They are the label of the binary variables.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : HOBO.
            Updates the HOBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = HOBO()
        >>> H.add_constraint_NOR('a', 'b')  # enforce a NOR b

        >>> H = HOBO()
        >>> H.add_constraint_NOR(['a', 'b'], ['c', 'd'])
        >>> # enforce (a AND b) NOR (c AND d)

        >>> H = HOBO()
        >>> H.add_constraint_NOR('a', 'b', 'c', 'd')
        >>> # enforce a NOR b NOR c NOR d

        >>> H = HOBO()
        >>> # enforce (a AND b) NOR (c AND d NAD e) NOR f
        >>> H.add_constraint_NOR(['a', 'b'], ['c', 'd', 'e'], 'f')

        """
        P = 1 - HOBO().add_constraint_OR(*variables)
        return self.add_constraint_eq_zero(P, lam=lam, bounds=(0, 1))

    def add_constraint_XNOR(self, *variables, lam=1):
        r"""add_constraint_XNOR.

        Add a penalty to the HOBO that is only nonzero when
        :math:`\lnot(v_0 \oplus v_1 \oplus ... \oplus v_n)` is True, with a
        penalty  factor ``lam``, where ``v_0 = variables[0]``,
        ``v_1 = variables[1]``, ..., ``v_n = variables[-1]``. See
        ``qubovert.sat.XNOR`` for the XNOR convention that qubovert uses for
        more than two inputs.

        Parameters
        ----------
        *variables : arguments.
            Each element of variables is a hashable object or a list of
            hashable objects. They are the label of the binary variables.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : HOBO.
            Updates the HOBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = HOBO()
        >>> H.add_constraint_XNOR('a', 'b')  # enforce a XNOR b

        >>> H = HOBO()
        >>> H.add_constraint_XNOR(['a', 'b'], ['c', 'd'])
        >>> # enforce (a AND b) XNOR (c AND d)

        >>> H = HOBO()
        >>> H.add_constraint_XNOR('a', 'b', 'c')  # enforce a XNOR b XNOR c

        """
        P = 1 - HOBO().add_constraint_XOR(*variables)
        return self.add_constraint_eq_zero(P, lam=lam, bounds=(0, 1))

    def add_constraint_NOT(self, a, lam=1):
        r"""add_constraint_NOT.

        Add a penalty to the HOBO that is only nonzero when
        :math:`\lnot a` is True, with a penalty factor ``lam``.

        Parameters
        ----------
        a : any hashable object or a list of hashable objects.
            The label for binary variables ``a``.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : HOBO.
            Updates the HOBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = HOBO()
        >>> H.add_constraint_NOT('a')  # enforce not a
        >>> H
        {('a',): 1}

        >>> H = HOBO()
        >>> H.add_constraint_NOT(['a', 'b'])  # enforce NOT (a AND b)
        >>> H
        {(a, b): 1}

        """
        P = PUBO({_create_tuple(a): 1})
        return self.add_constraint_eq_zero(P, lam=lam, bounds=(0, 1))



#   Copyright 2019 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""``qubovert`` is a module for converting problems into QUBO/Ising form.

QUBO stands for Quadratic Unconstrained Binary Optimization. QUBO problems
have a one-to-one mapping to classical Ising problems, and most optimization
problems are formatted in QUBO form when the solver is a quantum computer.
See ``qubovert.__all__`` for useful functionality, ``qubovert.problems__all__``
for problems defined, and ``qubovert.utils.__all__`` for some utility
functions, and ``qubovert.sat.__all__`` for the satisfiability library.

"""

from ._version import __version__

from . import utils

from ._qubo import *
from ._ising import *
from ._pubo import *
from ._hising import *
from ._hobo import *
from ._hoio import *

from ._qubo import __all__ as __all_qubo__
from ._ising import __all__ as __all_ising__
from ._pubo import __all__ as __all_pubo__
from ._hising import __all__ as __all_hising__
from ._hobo import __all__ as __all_hobo__
from ._hoio import __all__ as __all_hoio__

__all__ = (
    __all_qubo__ +
    __all_ising__ +
    __all_pubo__ +
    __all_hising__ +
    __all_hobo__ +
    __all_hoio__
)

from . import sat
from . import problems

name = "qubovert"



#   Copyright 2019 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_conversions.py.

This file contains methods to convert to and from QUBO/Ising/PUBO/HIsing
problems, as well as various misc converters

"""

from . import QUBOMatrix, IsingMatrix, PUBOMatrix, HIsingMatrix
# for QUBO, Ising, PUBO, HIsing, can't import directly because it will cause
# circular imports, so instead just import qubovert.
import qubovert


__all__ = (
    'qubo_to_ising', 'ising_to_qubo', 'pubo_to_hising', 'hising_to_pubo',
    'Conversions'
)


def qubo_to_ising(Q):
    """qubo_to_ising.

    Convert the specified QUBO problem into an Ising problem. Note that
    QUBO {0, 1} values go to Ising {-1, 1} values in that order!

    Parameters
    ----------
    Q : dictionary, qubovert.QUBO, or qubovert.utils.QUBOMatrix object.
        Maps tuples of binary variables indices to the Q value. See
        ``help(qubovert.QUBO)`` and ``help(qubovert.utils.QUBOMatrix)`` for
        info on formatting.

    Returns
    ------
    L : qubovert.Ising or qubovert.utils.IsingMatrix object.
        tuple of spin labels map to Ising values. If ``Q`` is a
        ``qubovert.utils.QUBOMatrix`` object, then ``L`` will be a
        ``qubovert.utils.IsingMatrix``, otherwise ``L`` will be a
        ``qubovert.Ising`` object.

    Example
    -------
    >>> Q = {(0,): 1, (0, 1): -1, (1,): 3}
    >>> L = qubo_to_ising(Q)
    >>> isinstance(L, qubovert.utils.IsingMatrix)
    True

    >>> Q = {('a',): 1, ('a', 'b'): -1, ('b',): 3}
    >>> L = qubo_to_ising(Q)
    >>> isinstance(L, qubovert.Ising)
    True

    """
    # could just use IsingMatrix(pubo_to_hising(Q)), but then we spend a lot of
    # time converting from a HIsingMatrix to IsingMatrix, so instead we
    # explictly write out the conversion.

    # not isinstance! because isinstance(QUBO, QUBOMatrix) is True
    if type(Q) == QUBOMatrix:
        L = IsingMatrix()
        squash_key = QUBOMatrix.squash_key
    else:
        L = qubovert.Ising()
        squash_key = qubovert.QUBO.squash_key

    for kp, v in Q.items():
        k = squash_key(kp)
        if not k:
            L[k] += v
        elif len(k) == 1:
            L[k] += v / 2
            L[()] += v / 2
        elif len(k) == 2:
            i, j = k
            L[k] += v / 4
            L[(i,)] += v / 4
            L[(j,)] += v / 4
            L[()] += v / 4
        # len(k) cannot be greater than 2 because the squash_key checks

    return L


def ising_to_qubo(L):
    """ising_to_qubo.

    Convert the specified Ising problem into an upper triangular QUBO problem.
    Note that Ising {-1, 1} values go to QUBO {0, 1} values in that order!

    Parameters
    ----------
    L : dictionary, qubovert.Ising, or qubovert.utils.IsingMatrix object.
        Tuple of spin labels map to Ising values. See
        ``help(qubovert.Ising)`` and ``help(qubovert.utils.IsingMatrix)`` for
        info on formatting.


    Returns
    -------
    Q : qubovert.QUBO or qubovert.utils.QUBOMatrix object.
        If ``L`` is a ``qubovert.utils.IsingMatrix`` object, then ``Q`` will be
        a ``qubovert.utils.QUBOMatrix``, otherwise ``Q`` will be a
        ``qubovert.QUBO`` object. See ``help(qubovert.QUBO)`` and
        ``help(qubovert.utils.QUBOMatrix)`` for info on formatting.

    Example
    -------
    >>> L = {(0,): 1, (1,): -1, (0, 1): -1}
    >>> Q = ising_to_qubo(L)
    >>> isinstance(Q, qubovert.utils.QUBOMatrix)
    True

    >>> L = {('a',): 1, ('b',): -1, (0, 1): -1}
    >>> Q = ising_to_qubo(L)
    >>> isinstance(Q, qubovert.QUBO)
    True

    """
    # could just use QUBOMatrix(hising_to_pubo(L)), but then we spend a lot of
    # time converting from a PUBOMatrix to QUBOMatrix, so instead we explictly
    # write out the conversion.

    # not isinstance! because isinstance(Ising, IsingMatrix) is True
    if type(L) == IsingMatrix:
        Q = QUBOMatrix()
        squash_key = IsingMatrix.squash_key
    else:
        Q = qubovert.QUBO()
        squash_key = qubovert.Ising.squash_key

    for kp, v in L.items():
        k = squash_key(kp)
        if not k:
            Q[k] += v
        elif len(k) == 1:
            Q[k] += 2 * v
            Q[()] -= v
        elif len(k) == 2:
            i, j = k
            Q[k] += 4 * v
            Q[(i,)] -= 2 * v
            Q[(j,)] -= 2 * v
            Q[()] += v
        # squash key ensures that len(k) <= 2

    return Q


def pubo_to_hising(P):
    """pubo_to_hising.

    Convert the specified PUBO problem into an HIsing problem. Note that
    PUBO {0, 1} values go to HIsing {-1, 1} values in that order!

    Parameters
    ----------
    P : dictionary, qubovert.PUBO, or qubovert.utils.PUBOMatrix object.
        Maps tuples of binary variables indices to the P value. See
        ``help(qubovert.PUBO)`` and ``help(qubovert.utils.PUBOMatrix)`` for
        info on formatting.

    Returns
    ------
    H : qubovert.utils.HIsingMatrix object.
        tuple of spin labels map to HIsing values. If ``P`` is a
        ``qubovert.utils.PUBOMatrix`` object, then ``H`` will be a
        ``qubovert.utils.HIsingMatrix``, otherwise ``H`` will be a
        ``qubovert.HIsing`` object..

    Example
    -------
    >>> P = {(0,): 1, (0, 1): -1, (1,): 3}
    >>> H = pubo_to_hising(P)
    >>> isinstance(H, qubovert.utils.HIsingMatrix)
    True

    >>> P = {(0,): 1, (0, 1): -1, (1,): 3}
    >>> H = pubo_to_hising(P)
    >>> isinstance(H, qubovert.Ising)
    True

    """
    def generate_new_key_value(k):
        """generate_new_key_value.

        Recursively generate the PUBO key, value pairs for converting the
        product ``x[k[0]] * ... * x[k[-1]]``, where each ``x`` is a binary
        variable in {0, 1}, to the product
        ``(z[k[0]]+1)/2 * ... * (z[k[1]]+1)/2``., where each ``z`` is a spin
        in {-1, 1}.

        Parameters
        ----------
        k : tuple.
            Each element of the tuple corresponds to a binary label.

        Yields
        ------
        res : tuple (key, value)
            key : tuple.
                Each element of the tuple corresponds to a spin label.
            value : float.
                The value to multiply the value corresponding with ``k`` by.

        """
        if not k:
            yield k, 1
        else:
            for key, value in generate_new_key_value(k[1:]):
                yield (k[0],) + key, value / 2
                yield key, value / 2

    # not isinstance! because isinstance(PUBO, PUBOMatrix) is True
    if type(P) == PUBOMatrix:
        H = HIsingMatrix()
        squash_key = PUBOMatrix.squash_key
    else:
        H = qubovert.HIsing()
        squash_key = qubovert.PUBO.squash_key

    for k, v in P.items():
        for key, value in generate_new_key_value(squash_key(k)):
            H[key] += value * v

    return H


def hising_to_pubo(H):
    """hising_to_pubo.

    Convert the specified HIsing problem into an upper triangular PUBO problem.
    Note that HIsing {-1, 1} values go to PUBO {0, 1} values in that order!

    Parameters
    ----------
    H : dictionary or qubovert.utils.HIsingMatrix object.
        Tuple of spin labels map to HIsing values. See
        ``help(qubovert.HIsing)`` and ``help(qubovert.utils.HIsingMatrix)`` for
        info on formatting.

    Returns
    -------
    P : qubovert.utils.PUBOMatrix object.
        If ``H`` is a ``qubovert.utils.HIsingMatrix`` object, then ``P`` will
        be a ``qubovert.utils.PUBOMatrix``, otherwise ``P`` will be a
        ``qubovert.PUBO`` object. See ``help(qubovert.PUBO)`` and
        ``help(qubovert.utils.PUBOMatrix)`` for info on formatting.

    Example
    -------
    >>> H = {(0,): 1, (1,): -1, (0, 1): -1}
    >>> P = hising_to_pubo(H)
    >>> isinstance(P, qubovert.utils.PUBOMatrix)
    True

    >>> H = {('0',): 1, ('1',): -1, (0, 1): -1}
    >>> P = hising_to_pubo(H)
    >>> isinstance(P, qubovert.PUBO)
    True

    """
    def generate_new_key_value(k):
        """generate_new_key_value.

        Recursively generate the PUBO key, value pairs for converting the
        product ``z[k[0]] * ... * z[k[-1]]``, where each ``z`` is a spin in
        {-1, 1}, to the product ``(2*x[k[0]]-1) * ... * (2*x[k[1]]-1)``, where
        each ``x`` is a binary variables in {0, 1}.

        Parameters
        ----------
        k : tuple.
            Each element of the tuple corresponds to a spin label.

        Yields
        ------
        res : tuple (key, value)
            key : tuple.
                Each element of the tuple corresponds to a binary label.
            value : float.
                The value to multiply the value corresponding with ``k`` by.

        """
        if not k:
            yield k, 1
        else:
            for key, value in generate_new_key_value(k[1:]):
                yield (k[0],) + key, 2 * value
                yield key, -value

    # not isinstance! because isinstance(Ising, IsingMatrix) is True
    if type(H) == HIsingMatrix:
        P = PUBOMatrix()
        squash_key = HIsingMatrix.squash_key
    else:
        P = qubovert.PUBO()
        squash_key = qubovert.HIsing.squash_key

    for k, v in H.items():
        for key, value in generate_new_key_value(squash_key(k)):
            P[key] += value * v

    return P


class Conversions:
    """Conversions.

    This is a parent class that defines the functions ``to_qubo``,
    ``to_ising``, ``to_pubo``, and ``to_hising``. Any subclass that inherits
    from ``Conversions`` `must` supply at least one of ``to_qubo`` or
    ``to_ising``. And at least one of ``to_pubo`` or ``to_hising``.


    """

    def to_qubo(self, *args, **kwargs):
        """to_qubo.

        Create and return upper triangular QUBO representing the problem.
        Should be implemented in child classes. If this method is not
        implemented in the child class, then it simply calls ``to_ising`` and
        converts the ising formulation to a QUBO formulation.

        Parameters
        ----------
        arguments : Defined in the child class.
            They should be parameters that define lagrange multipliers or
            factors in the QUBO.

        Return
        -------
        Q : qubovert.utils.QUBOMatrix object.
            The upper triangular QUBO matrix, a QUBOMatrix object.
            For most practical purposes, you can use QUBOMatrix in the
            same way as an ordinary dictionary. For more information,
            see ``help(qubovert.utils.QUBOMatrix)``.

        Raises
        -------
        ``RecursionError`` if neither ``to_qubo`` nor ``to_ising`` are defined
        in the subclass.

        """
        return ising_to_qubo(self.to_ising(*args, **kwargs))

    def to_ising(self, *args, **kwargs):
        """to_ising.

        Create and return Ising model representing the problem.
        Should be implemented in child classes. If this method is not
        implemented in the child class, then it simply calls ``to_qubo`` and
        converts the QUBO formulation to an Ising formulation.

        Parameters
        ----------
        arguments : Defined in the child class.
            They should be parameters that define lagrange multipliers or
            factors in the Ising model.

        Return
        ------
        L : qubovert.utils.IsingMatrix object.
            The upper triangular coupling matrix, where two element tuples
            represent couplings and one element tuples represent fields.
            For most practical purposes, you can use IsingCoupling in the
            same way as an ordinary dictionary. For more information,
            see ``help(qubovert.utils.IsingMatrix)``.

        Raises
        -------
        ``RecursionError`` if neither ``to_qubo`` nor ``to_ising`` are defined
        in the subclass.

        """
        return qubo_to_ising(self.to_qubo(*args, **kwargs))

    def to_pubo(self, *args, **kwargs):
        """to_pubo.

        Create and return upper triangular PUBO representing the problem.
        Should be implemented in child classes. If this method is not
        implemented in the child class, then it simply calls ``to_hising`` or
        ``to_qubo`` and converts the hising or QUBO formulations to a
        PUBO formulation.

        Parameters
        ----------
        arguments : Defined in the child class.
            They should be parameters that define lagrange multipliers or
            factors in the QUBO.

        Return
        -------
        P : qubovert.utils.PUBOMatrix object.
            The upper triangular PUBO matrix, a PUBOMatrix object.
            For most practical purposes, you can use PUBOMatrix in the
            same way as an ordinary dictionary. For more information,
            see ``help(qubovert.utils.PUBOMatrix)``.

        Raises
        -------
        ``RecursionError`` if neither ``to_pubo`` nor ``to_hising`` are defined
        in the subclass.

        """
        return hising_to_pubo(self.to_hising(*args, **kwargs))

    def to_hising(self, *args, **kwargs):
        """to_hising.

        Create and return HIsing model representing the problem.
        Should be implemented in child classes. If this method is not
        implemented in the child class, then it simply calls ``to_pubo`` or
        ``to_ising`` and converts to a HIsing formulation.

        Parameters
        ----------
        arguments : Defined in the child class.
            They should be parameters that define lagrange multipliers or
            factors in the Ising model.

        Return
        ------
        H : qubovert.utils.HIsingMatrix object.
            For most practical purposes, you can use HIsingMatrix in the
            same way as an ordinary dictionary. For more information,
            see ``help(qubovert.utils.HIsingMatrix)``.

        Raises
        -------
        ``RecursionError`` if neither ``to_pubo`` nor ``to_hising`` are defined
        in the subclass.

        """
        return pubo_to_hising(self.to_pubo(*args, **kwargs))



#   Copyright 2019 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""setup.py.

Set up details for ``pip install qubovert`` or ``pip install -e .`` if
installing by source.

"""

import setuptools
from qubovert import __version__, __name__


with open('README.rst') as f:
    README = f.read()

with open("requirements.txt") as f:
    REQUIREMENTS = [line.strip() for line in f if line.strip()]


setuptools.setup(
    name=__name__,
    version=__version__,
    author="Joseph T. Iosue",
    author_email="joe.iosue@yahoo.com",
    description="A package for converting common problems to QUBO/Ising form",
    long_description=README,
    long_description_content_type='text/x-rst',
    url="https://github.com/jiosue/qubovert",
    license="Apache Software License 2.0",
    packages=setuptools.find_packages(exclude=("tests", "docs")),
    test_suite="tests",
    install_requires=REQUIREMENTS,
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    project_urls={
        "Source": "https://github.com/jiosue/qubovert",
        "Docs": "https://qubovert.readthedocs.io"
    }
)
